from radical.entk import Pipeline, Stage, Task, AppManager
import os
from random import randint
# ------------------------------------------------------------------------------
# Set default verbosity
# if os.environ.get('RADICAL_ENTK_VERBOSE') == None:
    # os.environ['RADICAL_ENTK_REPORT'] = 'True'


# Description of how the RabbitMQ process is accessible
# No need to change/set any variables if you installed RabbitMQ has a system
# process. If you are running RabbitMQ under a docker container or another
# VM, set "RMQ_HOSTNAME" and "RMQ_PORT" in the session where you are running
# this script.
hostname = os.environ.get('RMQ_HOSTNAME', 'localhost')
port = int(os.environ.get('RMQ_PORT', 5672))


num_replicas = 8
pipelines = []
# Replica state
# 0 - Ready for MD
# 1 - MD done, waiting for exchange selection
# 2 - Selected for exchange
# 3 - Not selected for exchange
# 4 - Done
replica_state = [0 for _ in range(num_replicas)]
finish_threshold = 4

replica_cycles = [0 for _ in range(num_replicas)]
cycle_threshold = 4

replica_neighbors = dict()
for ind in range(num_replicas):
    replica_neighbors['R_%s'%ind] = list()


def generate_pipeline(index):

    def md_condition():

        print 'Replica %s | Cycle %s | Event: Evaluating condition after MD'%(index, replica_cycles[index])
        
        global replica_state
        replica_state[index] = 1

        if replica_state.count(1) == finish_threshold:            
            return True
            
        return False


    def md_on_true():

        print 'Replica %s | Cycle %s | Event: Adding exchange'%(index, replica_cycles[index])

        global replica_state
        
        replica_state[index] = 2
        s_exch = create_exch_stage()        
        p.add_stages(s_exch)
        exch_r_ids = list()
        for ind, state in enumerate(replica_state):
            if state==1:
                replica_state[ind] = 3
                replica_neighbors['R_%s'%index].append(ind)
                exch_r_ids.append(str(ind))

        print 'Replica %s | Cycle %s | Event: Initiating exchanges between replicas %s'%(index, replica_cycles[index], ' '.join(exch_r_ids))

    def md_on_false():

        print 'Replica %s | Cycle %s | Event: No exchange added'%(index, replica_cycles[index])


    def create_md_stage():

        print 'Replica %s | Cycle %s | Event: Creating MD task'%(index, replica_cycles[index])

        # Create a Stage object
        s_md = Stage()

        # Create a Task object which creates a file named 'output.txt' of size 1 MB
        t_md = Task()
        t_md.executable = ['/bin/sleep']
        t_md.arguments = ['%s'%randint(2,7)]

        # Add the Task to the Stage
        s_md.add_tasks(t_md)

        s_md.post_exec = {  'condition': md_condition,
                            'on_true': md_on_true,
                            'on_false': md_on_false
                        }

        return s_md


    def exch_condition():    

        print 'Replica %s | Cycle %s | Event: Evaluation condition after exchange'%(index, replica_cycles[index])

        global replica_state, replica_cycles

        # Always returns True
        for ind, state in enumerate(replica_state):
            if ind in replica_neighbors['R_%s'%index]:
                if state in [2,3]:
                    replica_cycles[ind] += 1
                    if replica_cycles[ind] == cycle_threshold:
                        print 'Replica %s completed'%ind
                        replica_state[ind] = 4

        return True

    def exch_on_true():

        print 'Replica %s | Cycle %s | Event: Adding MD tasks to replicas after exchange'%(index, replica_cycles[index])

        global replica_state, pipelines
        for ind, state in enumerate(replica_state):
            if ind in replica_neighbors['R_%s'%index]:
                if state in [2,3]:
                    s_md = create_md_stage()
                    pipelines[ind].add_stages(s_md)
                    print 'Replica %s | Cycle %s | Event: MD task added to replica %s'%(index, replica_cycles[index], ind)
                    pipelines[ind].rerun()
                    replica_state[ind] = 0

    def exch_on_false():

        pass

    def create_exch_stage():

        print 'Replica %s | Cycle %s | Event: Creating exchange task'%(index, replica_cycles[index])

        s_exch = Stage()
        t_exch = Task()
        t_exch.executable = ['/bin/sleep']
        t_exch.arguments =  ['5']

        s_exch.add_tasks(t_exch)

        s_exch.post_exec = {
                            'condition': exch_condition,
                            'on_true': exch_on_true,
                            'on_false': exch_on_false
                        }

        return s_exch

    # Create a Pipeline object
    p = Pipeline()
    s_md = create_md_stage()    
    p.add_stages(s_md)

    return p


if __name__ == '__main__':

    for cnt in range(num_replicas):
        pipelines.append(generate_pipeline(index=cnt))

    # Create Application Manager
    appman = AppManager(hostname=hostname, port=port)

    # Create a dictionary describe four mandatory keys:
    # resource, walltime, and cpus
    # resource is 'local.localhost' to execute locally
    res_dict = {

        'resource': 'local.localhost_anaconda',
        'walltime': 10,
        'cpus': num_replicas
    }

    # Assign resource request description to the Application Manager
    appman.resource_desc = res_dict

    # Assign the workflow as a set or list of Pipelines to the Application Manager
    # Note: The list order is not guaranteed to be preserved
    appman.workflow = set(pipelines)

    # Run the Application Manager
    appman.run()



__copyright__ = "Copyright 2017-2018, http://radical.rutgers.edu"
__author__ = "Vivek Balasubramanian <vivek.balasubramanian@rutgers.edu>"
__license__ = "MIT"

import radical.utils as ru
from radical.entk.exceptions import *
import threading
from multiprocessing import Process, Event
from radical.entk import states, Task
from radical.entk.utils.init_transition import transition
import time
import json
import pika
import os
import radical.pilot as rp
from task_processor import create_cud_from_task, create_task_from_cu
from ..base.task_manager import Base_TaskManager
import Queue
import traceback


class TaskManager(Base_TaskManager):

    """
    A Task Manager takes the responsibility of dispatching tasks it receives from a queue for execution on to
    the available resources using a runtime system. In this case, the runtime system being used RADICAL Pilot. Once
    the tasks have completed execution, they are pushed on to another queue for other components of EnTK to access.


    :arguments:
        :pending_queue: List of queue(s) with tasks ready to be executed. Currently, only one queue.
        :completed_queue: List of queue(s) with tasks that have finished execution. Currently, only one queue.
        :rmgr: ResourceManager object to be used to access the Pilot where the tasks can be submitted
        :mq_hostname: Name of the host where RabbitMQ is running
        :port: port at which rabbitMQ can be accessed

    Currently, EnTK is configured to work with one pending queue and one completed queue. In the future, the number of
    queues can be varied for different throughput requirements at the cost of additional Memory and CPU consumption.
    """

    def __init__(self, sid, pending_queue, completed_queue,
                 rmgr, mq_hostname, port):

        super(TaskManager, self).__init__(sid,
                                          pending_queue,
                                          completed_queue,
                                          rmgr,
                                          mq_hostname,
                                          port,
                                          rts='radical.pilot')

        self._umgr = None
        self._rts_runner = None
        self._rmq_ping_interval = os.getenv('RMQ_PING_INTERVAL', 10)
        self._logger.info('Created task manager object: %s' % self._uid)
        self._prof.prof('tmgr obj created', uid=self._uid)

    # ------------------------------------------------------------------------------------------------------------------
    # Private Methods
    # ------------------------------------------------------------------------------------------------------------------

    def _tmgr(self, uid, umgr, rmgr, logger, mq_hostname, port, pending_queue, completed_queue):
        """
        **Purpose**: This method has 3 purposes: Respond to a heartbeat thread indicating the live-ness of the RTS,
        receive tasks from the pending_queue, start a new thread that processes these tasks and submits to the RTS.
        
        It is important to separate the reception of the tasks from their processing due to RMQ/AMQP design. The
        channel (i.e., the thread that holds the channel) that is receiving msgs from the RMQ server needs to be
        non-blocking as it can interfere with the heartbeat intervals of RMQ. Processing a large number of tasks can
        considerable time and can block the communication channel. Hence, the two are separated.
        
        The new thread is responsible for pushing completed tasks (returned by the RTS) to the dequeueing queue.
        It also converts Tasks into CUDs and CUs into (partially described) Tasks. This conversion is necessary since 
        the current RTS is RADICAL Pilot. Once Tasks are recovered from a CU, they are then pushed to the
        completed_queue. At all state transititons, they are synced (blocking) with the AppManager in the master process.

        In addition the tmgr also receives heartbeat 'request' msgs from the heartbeat-request queue. It responds with a
        'response' message to the heartbeart-response queue.

        **Details**: The AppManager can re-invoke the tmgr process with this function if the execution of the workflow is
        still incomplete. There is also population of a dictionary, placeholder_dict, which stores the path of each of
        the tasks on the remote machine.
        """

        try:

            def heartbeat_response(mq_channel):

                try:

                    # Get request from heartbeat-req for heartbeat response
                    hb_method_frame, hb_props, hb_body = mq_channel.basic_get(
                        queue=self._hb_request_q)

                    if hb_body:

                        logger.info('Received heartbeat request')

                        mq_channel.basic_publish(exchange='',
                                                 routing_key=self._hb_response_q,
                                                 properties=pika.BasicProperties(
                                                     correlation_id=hb_props.correlation_id),
                                                 body='response')

                        logger.info('Sent heartbeat response')
                        mq_channel.basic_ack(
                            delivery_tag=hb_method_frame.delivery_tag)

                except Exception, ex:
                    logger.exception(
                        'Failed to respond to heartbeat request, error: %s' % ex)
                    raise

            local_prof = ru.Profiler(
                name='radical.entk.%s' % self._uid + '-proc', path=self._path)
            local_prof.prof('tmgr process started', uid=self._uid)
            logger.info('Task Manager process started')

            # Acquire a connection+channel to the rmq server
            mq_connection = pika.BlockingConnection(
                pika.ConnectionParameters(host=mq_hostname, port=port))
            mq_channel = mq_connection.channel()

            # Make sure the heartbeat response queue is empty
            mq_channel.queue_delete(queue=self._hb_response_q)
            mq_channel.queue_declare(queue=self._hb_response_q)

            # Queue for communication between threads of this process
            task_queue = Queue.Queue()

            # Start second thread to receive tasks and push to RTS
            self._rts_runner = threading.Thread(target=self._process_tasks,
                                                args=(task_queue,
                                                      rmgr,
                                                      logger,
                                                      mq_hostname,
                                                      port,
                                                      local_prof,
                                                      self._sid))
            self._rts_runner.start()

            local_prof.prof('tmgr infrastructure setup done', uid=uid)

            last = time.time()
            while not self._tmgr_terminate.is_set():

                try:

                    # Get tasks from the pending queue
                    method_frame, header_frame, body = mq_channel.basic_get(
                        queue=pending_queue[0])

                    if body:

                        body = json.loads(body)
                        task_queue.put(body)

                        mq_channel.basic_ack(
                            delivery_tag=method_frame.delivery_tag)

                    heartbeat_response(mq_channel)

                except Exception, ex:
                    logger.exception('Error in task execution: %s' % ex)
                    raise

        except KeyboardInterrupt:

            self._logger.error('Execution interrupted by user (you probably hit Ctrl+C), ' +
                               'trying to cancel tmgr process gracefully...')

        except Exception, ex:

            print traceback.format_exc()
            raise EnTKError(ex)

        finally:

            local_prof.prof('terminating tmgr process', uid=uid)

            if self._rts_runner:
                self._rts_runner.join()

            mq_connection.close()
            local_prof.close()

    def _process_tasks(self, task_queue, rmgr, logger, mq_hostname, port, local_prof, sid):


        '''
        **Purpose**: The new thread that gets spawned by the main tmgr process invokes this function. This
        function receives tasks from 'task_queue' and submits them to the RADICAL Pilot RTS.
        '''

        placeholder_dict = dict()

        def load_placeholder(task, rts_uid):

            parent_pipeline = str(task.parent_pipeline['name'])
            parent_stage = str(task.parent_stage['name'])

            if parent_pipeline not in placeholder_dict:
                placeholder_dict[parent_pipeline] = dict()

            if parent_stage not in placeholder_dict[parent_pipeline]:
                placeholder_dict[parent_pipeline][parent_stage] = dict()

            if None not in [parent_pipeline, parent_stage, task.name]:
                placeholder_dict[parent_pipeline][parent_stage][str(task.name)] = {'path': str(task.path),
                                                                                   'rts_uid': rts_uid}

        def unit_state_cb(unit, state):

            try:

                logger.debug('Unit %s in state %s' % (unit.uid, unit.state))

                if unit.state in rp.FINAL:

                    # Acquire a connection+channel to the rmq server
                    mq_connection = pika.BlockingConnection(
                        pika.ConnectionParameters(host=mq_hostname, port=port))
                    mq_channel = mq_connection.channel()

                    task = None
                    task = create_task_from_cu(unit, local_prof)

                    transition(obj=task,
                               obj_type='Task',
                               new_state=states.COMPLETED,
                               channel=mq_channel,
                               queue='%s-cb-to-sync' % sid,
                               profiler=local_prof,
                               logger=logger)

                    load_placeholder(task, unit.uid)

                    task_as_dict = json.dumps(task.to_dict())

                    mq_channel.basic_publish(exchange='',
                                             routing_key='%s-completedq-1' % sid,
                                             body=task_as_dict
                                             # properties=pika.BasicProperties(
                                             # make message persistent
                                             #    delivery_mode = 2,
                                             # )
                                             )

                    logger.info('Pushed task %s with state %s to completed queue %s-completedq-1' % (task.uid, task.state,
                                                                                                     sid))

                    mq_connection.close()

            except KeyboardInterrupt:
                logger.exception('Execution interrupted by user (you probably hit Ctrl+C), ' +
                                 'trying to exit callback thread gracefully...')

                raise KeyboardInterrupt

            except Exception, ex:
                logger.exception('Error in RP callback thread: %s' % ex)

        umgr = rp.UnitManager(session=rmgr._session)
        umgr.add_pilots(rmgr.pilot)
        umgr.register_callback(unit_state_cb)

        mq_connection = pika.BlockingConnection(
            pika.ConnectionParameters(host=mq_hostname, port=port))
        mq_channel = mq_connection.channel()

        try:

            while not self._tmgr_terminate.is_set():

                body = None

                try:
                    body = task_queue.get(block=True, timeout=10)
                except Queue.Empty:
                    # Ignore empty exception, we don't always have new tasks to run
                    pass

                if body:

                    task_queue.task_done()

                    bulk_tasks = list()
                    bulk_cuds = list()

                    for task in body:
                        t = Task()
                        t.from_dict(task)
                        bulk_tasks.append(t)
                        bulk_cuds.append(create_cud_from_task(
                            t, placeholder_dict, local_prof))

                        transition(obj=t,
                                   obj_type='Task',
                                   new_state=states.SUBMITTING,
                                   channel=mq_channel,
                                   queue='%s-tmgr-to-sync' % sid,
                                   profiler=local_prof,
                                   logger=logger)

                    umgr.submit_units(bulk_cuds)

                    for task in bulk_tasks:

                        transition(obj=task,
                                   obj_type='Task',
                                   new_state=states.SUBMITTED,
                                   channel=mq_channel,
                                   queue='%s-tmgr-to-sync' % sid,
                                   profiler=local_prof,
                                   logger=logger)

        except KeyboardInterrupt as ex:
            logger.error('Execution interrupted by user (you probably hit Ctrl+C), ' +
                         'trying to cancel task processor gracefully...')

        except Exception as ex:
            print traceback.format_exc()
            raise EnTKError(ex)

    # ------------------------------------------------------------------------------------------------------------------
    # Public Methods
    # ------------------------------------------------------------------------------------------------------------------

    def start_manager(self):
        """
        **Purpose**: Method to start the tmgr process. The tmgr function
        is not to be accessed directly. The function is started in a separate
        thread using this method.
        """

        if not self._tmgr_process:

            try:

                self._prof.prof('creating tmgr process', uid=self._uid)
                self._tmgr_terminate = Event()

                self._tmgr_process = Process(target=self._tmgr,
                                             name='task-manager',
                                             args=(
                                                 self._uid,
                                                 self._umgr,
                                                 self._rmgr,
                                                 self._logger,
                                                 self._mq_hostname,
                                                 self._port,
                                                 self._pending_queue,
                                                 self._completed_queue)
                                             )

                self._logger.info('Starting task manager process')
                self._prof.prof('starting tmgr process', uid=self._uid)
                self._tmgr_process.start()

                return True

            except Exception, ex:

                self._logger.error('Task manager not started, error: %s' % ex)
                self.terminate_manager()
                raise

        else:
            self._logger.warn(
                'tmgr process already running, but attempted to restart!')
    # ------------------------------------------------------------------------------------------------------------------



__copyright__ = "Copyright 2017-2018, http://radical.rutgers.edu"
__author__ = "Vivek Balasubramanian <vivek.balasubramaniana@rutgers.edu>"
__license__ = "MIT"

import radical.utils as ru
from radical.entk.exceptions import EnTKError
import threading
from multiprocessing import Process, Event
from radical.entk import states, Task
import time
import json
import pika
import traceback
import os
import uuid
from ..base.task_manager import Base_TaskManager
from radical.entk.utils.init_transition import transition
import Queue


class TaskManager(Base_TaskManager):

    """
    A Task Manager takes the responsibility of dispatching tasks it receives from a pending_queue for execution on to
    the available resources using a runtime system. Once the tasks have completed execution, they are pushed on to
    the completed_queue for other components of EnTK to process.

    :arguments:
        :pending_queue: List of queue(s) with tasks ready to be executed. Currently, only one queue.
        :completed_queue: List of queue(s) with tasks that have finished execution. Currently, only one queue.
        :rmgr: ResourceManager object to be used to access the Pilot where the tasks can be submitted
        :mq_hostname: Name of the host where RabbitMQ is running
        :port: port at which rabbitMQ can be accessed

    Currently, EnTK is configured to work with one pending queue and one completed queue. In the future, the number of
    queues can be varied for different throughput requirements at the cost of additional Memory and CPU consumption.
    """

    def __init__(self, sid, pending_queue, completed_queue,
                 rmgr, mq_hostname, port):

        super(TaskManager, self).__init__(sid,
                                          pending_queue,
                                          completed_queue,
                                          rmgr,
                                          mq_hostname,
                                          port,
                                          rts='mock')

        self._rmq_ping_interval = os.getenv('RMQ_PING_INTERVAL', 10)

        self._logger.info('Created task manager object: %s' % self._uid)
        self._prof.prof('tmgr obj created', uid=self._uid)

    # ------------------------------------------------------------------------------------------------------------------
    # Private Methods
    # ------------------------------------------------------------------------------------------------------------------

    def _tmgr(self, uid, rmgr, logger, mq_hostname, port, pending_queue, completed_queue):
        """
        **Purpose**: Method to be run by the tmgr process. This method receives a Task from the pending_queue
        and submits it to the RTS. Currently, it also converts Tasks into CUDs and CUs into (partially described) Tasks.
        This conversion is necessary since the current RTS is RADICAL Pilot. Once Tasks are recovered from a CU, they
        are then pushed to the completed_queue. At all state transititons, they are synced (blocking) with the AppManager
        in the master process.

        In addition the tmgr also receives heartbeat 'request' msgs from the heartbeat-req queue. It responds with a
        'response' message to the 'heartbeart-res' queue.

        **Details**: The AppManager can re-invoke the tmgr process with this function if the execution of the workflow is
        still incomplete. There is also population of a dictionary, placeholder_dict, which stores the path of each of
        the tasks on the remote machine.
        """

        try:

            def heartbeat_response(mq_channel):

                try:

                    # Get request from heartbeat-req for heartbeat response
                    hb_method_frame, hb_props, hb_body = mq_channel.basic_get(
                        queue=self._hb_request_q)

                    if hb_body:

                        logger.info('Received heartbeat request')

                        mq_channel.basic_publish(exchange='',
                                                 routing_key=self._hb_response_q,
                                                 properties=pika.BasicProperties(
                                                     correlation_id=hb_props.correlation_id),
                                                 body='response')

                        logger.info('Sent heartbeat response')
                        mq_channel.basic_ack(
                            delivery_tag=hb_method_frame.delivery_tag)

                except Exception, ex:
                    logger.exception(
                        'Failed to respond to heartbeat request, error: %s' % ex)
                    raise

            local_prof = ru.Profiler(
                name='radical.entk.%s' % self._uid + '-proc', path=self._path)

            local_prof.prof('tmgr process started', uid=self._uid)
            logger.info('Task Manager process started')

            # Thread should run till terminate condtion is encountered
            mq_connection = pika.BlockingConnection(
                pika.ConnectionParameters(host=mq_hostname, port=port))
            mq_channel = mq_connection.channel()

            # Queue for communication between threads of this process
            task_queue = Queue.Queue()

            # Start second thread to receive tasks and push to RTS
            self._rts_runner = threading.Thread(target=self._process_tasks,
                                                args=(task_queue,
                                                      rmgr,
                                                      logger,
                                                      mq_hostname,
                                                      port,
                                                      local_prof,
                                                      self._sid))
            self._rts_runner.start()

            local_prof.prof('tmgr infrastructure setup done', uid=uid)

            last = time.time()
            while not self._tmgr_terminate.is_set():

                try:

                    method_frame, header_frame, body = mq_channel.basic_get(
                        queue=pending_queue[0])

                    if body:

                        body = json.loads(body)
                        task_queue.put(body)

                        mq_channel.basic_ack(
                            delivery_tag=method_frame.delivery_tag)

                    heartbeat_response(mq_channel)

                except Exception, ex:
                    logger.exception('Error in task execution: %s' % ex)
                    raise

        except KeyboardInterrupt:

            self._logger.error('Execution interrupted by user (you probably hit Ctrl+C), ' +
                               'trying to cancel tmgr process gracefully...')

        except Exception, ex:

            print traceback.format_exc()
            raise EnTKError(ex)

        finally:

            local_prof.prof('terminating tmgr process', uid=uid)

            if self._rts_runner:
                self._rts_runner.join()

            mq_connection.close()
            local_prof.close()

    def _process_tasks(self, task_queue, rmgr, logger, mq_hostname, port, local_prof, sid):

        def load_placeholder(task):

            parent_pipeline = str(task.parent_pipeline['name'])
            parent_stage = str(task.parent_stage['name'])

            if parent_pipeline not in placeholder_dict:
                placeholder_dict[parent_pipeline] = dict()

            if parent_stage not in placeholder_dict[parent_pipeline]:
                placeholder_dict[parent_pipeline][parent_stage] = dict()

            if None not in [parent_pipeline, parent_stage, task.name]:
                placeholder_dict[parent_pipeline][parent_stage][str(
                    task.name)] = str(task.path)

        placeholder_dict = dict()

        mq_connection = pika.BlockingConnection(
            pika.ConnectionParameters(host=mq_hostname, port=port))
        mq_channel = mq_connection.channel()

        try:

            while not self._tmgr_terminate.is_set():

                body = None

                try:
                    body = task_queue.get(block=True, timeout=10)
                except Queue.Empty:
                    # Ignore empty exception, we don't always have new tasks to run
                    pass

                if body:

                    task_queue.task_done()

                    bulk_tasks = list()
                    bulk_cuds = list()

                    for task in body:
                        t = Task()
                        t.from_dict(task)
                        bulk_tasks.append(t)

                        transition(obj=t,
                                   obj_type='Task',
                                   new_state=states.SUBMITTING,
                                   channel=mq_channel,
                                   queue='%s-tmgr-to-sync' % self._sid,
                                   profiler=local_prof,
                                   logger=self._logger)

                    for task in bulk_tasks:

                        transition(obj=task,
                                   obj_type='Task',
                                   new_state=states.SUBMITTED,
                                   channel=mq_channel,
                                   queue='%s-tmgr-to-sync' % self._sid,
                                   profiler=local_prof,
                                   logger=self._logger)
                        self._logger.info(
                            'Task %s submitted to RTS' % (task.uid))

                    for task in bulk_tasks:

                        transition(obj=task,
                                   obj_type='Task',
                                   new_state=states.COMPLETED,
                                   channel=mq_channel,
                                   queue='%s-cb-to-sync' % self._sid,
                                   profiler=local_prof,
                                   logger=logger)

                        task_as_dict = json.dumps(task.to_dict())

                        mq_channel.basic_publish(exchange='',
                                                 routing_key='%s-completedq-1' % self._sid,
                                                 body=task_as_dict
                                                 # properties=pika.BasicProperties(
                                                 # make message persistent
                                                 #    delivery_mode = 2,
                                                 # )
                                                 )

                        logger.info('Pushed task %s with state %s to completed queue %s-completedq-1' % (
                            task.uid,
                            task.state,
                            sid))

        except KeyboardInterrupt as ex:
            logger.error('Execution interrupted by user (you probably hit Ctrl+C), ' +
                         'trying to cancel task processor gracefully...')

        except Exception as ex:
            print traceback.format_exc()
            raise EnTKError(ex)
        # ------------------------------------------------------------------------------------------------------------------
        # Public Methods
        # ------------------------------------------------------------------------------------------------------------------

    def start_manager(self):
        """
        **Purpose**: Method to start the tmgr process. The tmgr function
        is not to be accessed directly. The function is started in a separate
        thread using this method.
        """

        if not self._tmgr_process:

            try:

                self._prof.prof('creating tmgr process', uid=self._uid)
                self._tmgr_terminate = Event()

                self._tmgr_process = Process(target=self._tmgr,
                                             name='task-manager',
                                             args=(
                                                 self._uid,
                                                 self._rmgr,
                                                 self._logger,
                                                 self._mq_hostname,
                                                 self._port,
                                                 self._pending_queue,
                                                 self._completed_queue)
                                             )

                self._logger.info('Starting task manager process')
                self._prof.prof('starting tmgr process', uid=self._uid)
                self._tmgr_process.start()

                return True

            except Exception, ex:

                self._logger.error('Task manager not started, error: %s' % ex)
                self.terminate_manager()
                raise

        else:
            self._logger.warn(
                'tmgr process already running, but attempted to restart!')

    # ------------------------------------------------------------------------------------------------------------------



from radical.entk import Pipeline, Stage, Task, AppManager
import os
import sys
from time import sleep

# ------------------------------------------------------------------------------
# Set default verbosity

if not os.environ.get('RADICAL_ENTK_VERBOSE'):
    os.environ['RADICAL_ENTK_VERBOSE'] = 'INFO'

hostname = os.environ.get('RMQ_HOSTNAME', 'localhost')
port = int(os.environ.get('RMQ_PORT', 5672))
cur_dir = os.path.dirname(os.path.abspath(__file__))
MLAB = os.environ.get('RADICAL_PILOT_DBURL')


def generate_pipeline():

    def func_condition():

        p.suspend()
        print 'Suspending pipeline %s for 10 seconds' %p.uid
        sleep(10)
        return True

    def func_on_true():

        print 'Resuming pipeline %s' %p.uid
        p.resume()

    def func_on_false():
        pass

    # Create a Pipeline object
    p = Pipeline()

    # Create a Stage object
    s1 = Stage()

    for i in range(10):

        t1 = Task()
        t1.executable = ['sleep']
        t1.arguments = ['30']

        # Add the Task to the Stage
        s1.add_tasks(t1)

    # Add post-exec to the Stage
    s1.post_exec = {
        'condition': func_condition,
        'on_true': func_on_true,
        'on_false': func_on_false
    }

    # Add Stage to the Pipeline
    p.add_stages(s1)

    return p


if __name__ == '__main__':

    # Create a dictionary describe four mandatory keys:
    # resource, walltime, cores and project
    # resource is 'local.localhost' to execute locally
    res_dict = {

        'resource': 'local.localhost',
        'walltime': 15,
        'cpus': 2,
    }

    # Create Application Manager
    appman = AppManager(hostname=hostname, port=port)
    appman.resource_desc = res_dict

    p = generate_pipeline()

    # Assign the workflow as a set of Pipelines to the Application Manager
    appman.workflow = [p]

    # Run the Application Manager
    appman.run()
